@@ -2,7 +2,7 @@
 #define VIENNACL_LINALG_CG_HPP_
 
 /* =========================================================================
-   Copyright (c) 2010-2015, Institute for Microelectronics,
+   Copyright (c) 2010-2016, Institute for Microelectronics,
                             Institute for Analysis and Scientific Computing,
                             TU Wien.
    Portions of this software are copyright by UChicago Argonne, LLC.
@@ -38,6 +38,7 @@
 #include "viennacl/meta/result_of.hpp"
 #include "viennacl/linalg/iterative_operations.hpp"
 
+#include <Rcpp.h>
 namespace viennacl
 {
 namespace linalg
@@ -283,7 +284,7 @@ namespace detail
     CPU_NumericType norm_rhs_squared = ip_rr;
     CPU_NumericType new_ipp_rr_over_norm_rhs;
 
-    if (norm_rhs_squared <= tag.abs_tolerance()) //solution is zero if RHS norm is zero
+    if (norm_rhs_squared <= tag.abs_tolerance() * tag.abs_tolerance()) //solution is zero if RHS norm (squared) is zero
       return result;
 
     for (unsigned int i = 0; i < tag.max_iterations(); ++i)
